@@ -1,8 +1,11 @@
 /* libmii.c: MII diagnostic and setup library.
 
-	Copyright 1997-2001 by Donald Becker.
+	Copyright 1997-2003 by Donald Becker.
 	This version released under the Gnu General Public License,
 	incorporated herein by reference.
+	This source code may be distributed without modification using the
+	existing notice.  Any modification to this source code must include a
+	full notice as described in the GPL
 	Contact the author for use under other terms.
 
 	The author may be reached as becker@scyld.com, or C/O
@@ -15,8 +18,8 @@
 	http://www.national.com/pf/DP/DP83840A.html
 */
 
-static const char version[] =
-"libmii.c:v2.04 5/16/2001  Donald Becker (becker@scyld.com)\n"
+static const char version_msg[] =
+"libmii.c:v2.10 4/22/2003  Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 
 /* This library exports the following functions:
@@ -53,6 +56,10 @@ static void amd_tx(long ioaddr, int phy_id);
 static void admtek(long ioaddr, int phy_id);
 static void lu3x31(long ioaddr, int phy_id);
 static void myson981(long ioaddr, int phy_id);
+static void via_tahoe(long ioaddr, int phy_id);
+static void via_vt6103(long ioaddr, int phy_id);
+static void via_vt6105(long ioaddr, int phy_id);
+static void intel(long ioaddr, int phy_id);
 
 struct mii_partnum {
 	const char *vendor;				/* Vendor name. */
@@ -74,8 +81,9 @@ struct mii_partnum {
 	{"Quality Semiconductor QS6612", 0x0181, 0x4410, {0,}, 0, qs6612},
 	{"Quality Semiconductor (unknown type)", 0,0, {0x00, 0x60, 0x51}, 0, NULL},
 	{"SMSC 83c180", 0x0282, 0x1C51, {0}, 0, smsc83c180, },
-	{"TDK Semiconductor 78Q2120", 0x0300, 0xE542, {0,}, 0, tdk78q2120, },
-	{"TDK Semiconductor 78Q2120", 0x0300, 0xE543, {0,}, 0, tdk78q2120, },
+	{"TDK Semi 78Q2120 rev. 2", 0x0300, 0xE542, {0,}, 0, tdk78q2120, },
+	{"TDK Semi 78Q2120 rev. 3", 0x0300, 0xE543, {0,}, 0, tdk78q2120, },
+	{"TDK Semi 78Q2120 rev. 11", 0x0300, 0xE54B, {0,}, 0, tdk78q2120, },
 	{"TDK transceiver (unknown type)", 0,0, {0x00, 0xc0, 0x39}, 0, tdk78q2120},
 	{"Intel (unknown type)", 0,0, {0x00, 0xf8, 0x00}, 0, intel_i553},
 	{"Enable Semiconductor EL40-331", 0x0043, 0x7411, {0,}, 0, enablesemi},
@@ -95,6 +103,15 @@ struct mii_partnum {
 	{"LSI Logic (Seeq) 80225", 0, 0, {0x00,0xA0,0x7D}, 0, NULL},
 	{"Myson MTD981", 0x0302, 0xD000, {0,}, 0, myson981},
 	{"Myson (unknown type)", 0, 0, {0x00,0xC0,0xB4,}, 0, myson981},
+	{"Alta/Kendin Sundance", 0x0022, 0x1720, {0,}, 0, NULL},
+	{"Alta/Kendin Sundance", 0, 0, {0x00,0x08,0x85}, 0, NULL},
+	{"VIA Tahoe VT6103", 0x0101, 0x8f20, {0,}, 0, via_vt6103},
+	{"VIA Tahoe VT6104", 0x0101, 0x8f30, {0,}, 0, via_tahoe},
+	{"VIA Rhine VT6105", 0x0101, 0x8f22, {0,}, 0, via_vt6105},
+	{"Intel 82557 series", 0x02a8, 0x0150, {0,}, 0, intel},
+	{"Intel 82555 rev 1", 0x02a8, 0x0151, {0,}, 0, intel},
+	{"Intel 82559 transceiver", 0x02a8, 0x0154, {0,}, 0, intel},
+	{"Intel 82555 series transceiver", 0,0, {0x00,0xaa,0x00}, 0, intel},
 	{0, },
 };
 
@@ -105,6 +122,8 @@ void show_mii_details(long ioaddr, int phy_id)
 	int mii_reg, i, vendor = 0;
 	u16 bmcr, bmsr, new_bmsr;
 
+	/* This may not be omitted from the output. */
+	printf("%s", version_msg);
 	printf(" MII PHY #%d transceiver registers:", phy_id);
 	for (mii_reg = 0; mii_reg < 32; mii_reg++) {
 		mii_val[mii_reg] = mdio_read(ioaddr, phy_id, mii_reg);
@@ -299,6 +318,15 @@ static void msg_if_set(const int val, const struct msg_tbl msg_tbl[])
 			printf(" %s\n", msg_tbl[i].msg);
 }
 
+static void msg_if_set_fmt(const int val, const struct msg_tbl msg_tbl[],
+						   const char *fmt)
+{
+	int i;
+	for (i = 0; msg_tbl[i].bitmask; i++)
+		if (msg_tbl[i].bitmask & val)
+			printf(fmt, msg_tbl[i].msg);
+}
+
 static void qs6612(long ioaddr, int phy_id)
 {
 	printf("  QS6612 extra registers: Mode %4.4x.\n"
@@ -480,7 +508,7 @@ static void lu3x31(long ioaddr, int phy_id)
 
 static const struct msg_tbl myson_reg16[] = {
 	{0x0080, " Far end fault enabled."},
-	{0x0040, " Transformer ration 1.25:1."},
+	{0x0040, " Transformer ratio 1.25:1."},
 	{0x0020, " Polarity correction diabled."},
 	{0x0010, " Link is forced up regardless of link beat."},
 	{0x0004, " Bypass Jabber check."},
@@ -515,6 +543,78 @@ static void myson981(long ioaddr, int phy_id)
 	return;
 }
 
+/* These are much like the TDK events in reversed order. */
+static const struct msg_tbl via_reg17[] = {
+	{0x0001, "Auto-Negotiation complete"},
+	{0x0002, "Remote fault detected"},
+	{0x0004, "Link failure detected"},
+	{0x0008, "Bad Start Stream detected"},
+	{0x0010, "Parallel detection fault"},
+	{0x0020, "Extended negotiation page received"},
+	{0x0040, "5B/4B code error detected"},
+	{0x0080, "Jabber detected"},
+	{0, 0},
+};
+
+static void via_tahoe(long ioaddr, int phy_id)
+{
+	int mii_reg16 = mii_val[16];
+	int mii_reg17 = mii_val[17];
+	int mii_reg18 = mii_val[18];
+
+	printf("  VIA Tahoe extended registers: 16 %4.4x  #17 %4.4x  #18 %4.4x.\n",
+		   mii_reg16, mii_reg17, mii_reg18);
+	msg_if_set_fmt(mii_reg17, via_reg17, "   %s\n");
+	printf("   Link %s 10%s Mbps %s duplex\n",
+		   mii_reg18 & 0x2000 ? "up" : "down",
+		   mii_reg18 & 0x0400 ? "0" : "",
+		   mii_reg18 & 0x0800 ? "full" : "half");
+}
+
+/* Information from
+   http://www.via.com.tw/en/datasheet/DS6103110.pdf
+*/
+   
+static void via_vt6103(long ioaddr, int phy_id)
+{
+	printf("  VIA vt6103 error counts since the last check:\n"
+		   "   The link has failed %d times.\n"
+		   "   The receiver has lost lock %d times.\n"
+		   "   There have been %d false carrier/SQE error.\n",
+		   mii_val[21], mii_val[22], mii_val[23]);
+}
+
+/* Information from
+   http://www.via.com.tw/en/Networking/DS6105LOM100.pdf
+*/
+
+static void via_vt6105(long ioaddr, int phy_id)
+{
+	printf("  VIA vt6105 PHY status:\n"
+		   "   Duplex %s  speed %s\n",
+		   mii_val[20] & 0x0001 ? "full" : "half",
+		   mii_val[20] & 0x0002 ? "100" : "10");
+}
+
+/* Information from
+   http://www.via.com.tw/en/Networking/DS6105LOM100.pdf
+*/
+static void intel(long ioaddr, int phy_id)
+{
+	printf("  Intel 8255* PHY #%d extended management registers:\n"
+		   "    Error counts, cleared when read:\n"
+		   "     False carriers %d\n"
+		   "     Link disconnects %d\n"
+		   "     Receive errors %d\n"
+		   "     Rx symbol errors %d.\n"
+		   "     Rx 10Mbps Early End-Of-Frame errors %d.\n"
+		   "     Rx 100Mbps Early End-Of-Frame errors %d.\n"
+		   "     Tx jabber errors %d.\n",
+		   mii_val[18],
+		   mii_val[19], mii_val[20], mii_val[21], mii_val[22], mii_val[23],
+		   mii_val[24], mii_val[25]);
+}
+
 
 /*
  * Local variables:
