@@ -3,11 +3,17 @@
 
 	Usage:	mii-diag [-vw] interface.
 
+	This program reads and writes the Media Independent Interface (MII)
+	management registers on network transceivers.  The registers control
+	and report network link settings and errors.  Examples are link speed,
+	duplex, capabilities advertised to the link partner, status LED
+	indications and link error counters.
+
 	Notes:
 	The compile-command is at the end of this source file.
-	This program only works with drivers that implement MII ioctl() calls.
+	This program works with drivers that implement MII ioctl() calls.
 
-	Written/copyright 1997-2001 by Donald Becker <becker@scyld.com>
+	Written/copyright 1997-2003 by Donald Becker <becker@scyld.com>
 
 	This program is free software; you can redistribute it
 	and/or modify it under the terms of the GNU General Public
@@ -15,7 +21,7 @@
 
 	The author may be reached as becker@scyld.com, or C/O
 	 Scyld Computing Corporation
-	 410 Severn Ave., Suite 210
+	 914 Bay Ridge Road, Suite 220
 	 Annapolis MD 21403
 
 	References
@@ -25,49 +31,51 @@
 */
 
 static char version[] =
-"mii-diag.c:v2.02 5/21/2001 Donald Becker (becker@scyld.com)\n"
+"mii-diag.c:v2.09 9/06/2003 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 
 static const char usage_msg[] =
-"Usage: mii-diag [-aDfrRvVwh] [-AF <to-advertise>] [--watch] <interface>.\n
-               -h -? --help
-                       Show summary options
-               -v
-                       Show version
-               -V --verbose
-                       Verbose mode
-               -D --debug
-                       Debug mode
-               -a --all
-                       Show all interfaces
-               -A --Advertise <Mode>
-                       Change the capabilities advertised by your network
-                       interface. Currently the following modes are allowed:
-                       100baseT4, 100baseTx, 100baseTx-FD, 100baseTx-HD,
-                       10baseT, 10baseT-FD, 10baseT-HD
-               -f --force
-                       Force the operation
-               -F --fixedspeed
-                       Force the network interface to a speed mode. Currently
-                       the following modes are allowed: 100baseT4, 100baseTx,
-                       100baseTx-FD, 100baseTx-HD, 10baseT, 10baseT-FD, 
-                       10baseT-HD
-               -G --parameters
-                       Write general settings value. Only use this, if you 
-                       know what you are doing!
-               -p --phy
-                       Set the PHY (MII address) to from which mii-diag
-                       generates the report.
-               -r -restart
-                       Restart the link negotiation.
-               -R -Reset
-                       Reset the tranceiver.
-	       -s --status
-	               Do return non-zero if no link beat is found.
-               -w --watch <interface>
-                       Constantly monitor te port and print out any changes.
-";
-
+"Usage: %s [--help] [-aDfrRvVw] [-AF <speed+duplex>] [--watch] <interface>.\n";
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [--watch] <interface>.\n"
+"\n"
+"  This program configures and monitors the transceiver management registers\n"
+"  for network interfaces.  It uses the Media Independent Interface (MII)\n"
+"  standard with additional Linux-specific controls to communicate with the\n"
+"  underlying device driver.  The MII registers control and report network\n"
+"  link settings and errors.  Examples are link speed, duplex, capabilities\n"
+"  advertised to the link partner, status LED indications and link error\n"
+"  counters.\n"
+"\n"
+"   The common usage is\n"
+"      mii-diag eth0\n"
+"\n"
+"   The default interface is \"eth0\".\n"
+" Frequently used options are\n"
+"   -A  --advertise <speed|setting>\n"
+"   -F  --fixed-speed <speed>\n"
+"	Speed is one of: 100baseT4, 100baseTx, 100baseTx-FD, 100baseTx-HD,\n"
+"	                 10baseT, 10baseT-FD, 10baseT-HD\n"
+"   -s  --status     Return exit status 2 if there is no link beat.\n"
+"\n"
+" Less frequently used options are\n"
+"   -a  --all-interfaces  Show the status all interfaces\n"
+"              (Not recommended with options that change settings.)\n"
+"   -D  --debug\n"
+"   -g  --read-parameters 	Get driver-specific parameters.\n"
+"   -G  --set-parameters PARMS	Set driver-specific parameters.\n"
+"       Parameters are comma separated, missing elements retain existing value.\n"
+"   -M  --msg-level LEVEL 	Set the driver message bit map.\n"
+"   -p  --phy ADDR		Set the PHY (MII address) to report.\n"
+"   -r  --restart	Restart the link autonegotiation.\n"
+"   -R  --reset		Reset the transceiver.\n"
+"   -v  --verbose	Report each action taken.\n"
+"   -V  --version	Emit version information.\n"
+"   -w  --watch		Continuously monitor the transceiver and report changes.\n"
+"\n"
+"   This command returns success (zero) if the interface information can be\n"
+"   read.  If the --status option is passed, a zero return means that the\n"
+"   interface has link beat.\n";
 
 #include <unistd.h>
 #include <stdlib.h>
@@ -89,29 +97,35 @@ static const char usage_msg[] =
 typedef unsigned short u16;
 typedef unsigned int u32;
 
-#ifndef SIOCGMIIPHY
-#define SIOCGMIIPHY (SIOCDEVPRIVATE)		/* Get the PHY in use. */
-#define SIOCGMIIREG (SIOCDEVPRIVATE+1) 		/* Read a PHY register. */
-#define SIOCSMIIREG (SIOCDEVPRIVATE+2) 		/* Write a PHY register. */
+#if defined(SIOCGPARAMS)  && SIOCGPARAMS != SIOCDEVPRIVATE+3
+#error Changed definition for SIOCGPARAMS
+#else
 #define SIOCGPARAMS (SIOCDEVPRIVATE+3) 		/* Read operational parameters. */
 #define SIOCSPARAMS (SIOCDEVPRIVATE+4) 		/* Set operational parameters. */
 #endif
 
+const char shortopts[] = "aA:C:DfF:gG:hmM:p:rRsvVw?";
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
     {"all-interfaces", 0, 0, 'a'},	/* Show all interfaces. */
-	{"Advertise",	1, 0, 'A'},		/* Change the capabilities advertised. */
+	{"advertise",	1, 0, 'A'},		/* Change the capabilities advertised. */
+	{"BMCR",		1, 0, 'C'},		/* Set the control register. */
     {"debug",       0, 0, 'D'},		/* Increase the debug level. */
     {"force",       0, 0, 'f'},		/* Force the operation. */
-    {"parameters",  0, 0, 'G'},		/* Write general settings value. */
+    {"fixed-speed", 1, 0, 'F'},		/* Fixed speed name. */
+    {"read-parameters", 0, 0, 'g'}, /* Show general settings values. */
+    {"set-parameters",  1, 0, 'G'},	/* Write general settings values. */
+    {"help", 		0, 0, 'h'},		/* Print a long usage message. */
+    {"monitor",		0, 0, 'm'},		/* Monitor status register. */
+    {"msg-level",	1, 0, 'M'},		/* Set the driver message level. */
     {"phy",			1, 0, 'p'},		/* Set the PHY (MII address) to report. */
     {"restart",		0, 0, 'r'},		/* Restart the link negotiation */
-    {"Reset",		0, 0, 'R'},		/* Reset the transceiver. */
+    {"reset",		0, 0, 'R'},		/* Reset the transceiver. */
     {"status",		0, 0, 's'},		/* Non-zero exit status w/ no link beat. */
-    {"help", 		0, 0, '?'},		/* Give help */
     {"verbose", 	0, 0, 'v'},		/* Report each action taken.  */
     {"version", 	0, 0, 'V'},		/* Emit version information.  */
     {"watch", 		0, 0, 'w'},		/* Constantly monitor the port.  */
+    {"error", 		0, 0, '?'},		/* Return the error message. */
     { 0, 0, 0, 0 }
 };
 
@@ -122,7 +136,8 @@ extern void monitor_mii(long ioaddr, int phy_id);
 /* Command-line flags. */
 unsigned int opt_a = 0,					/* Show-all-interfaces flag. */
 	opt_f = 0,					/* Force the operation. */
-	opt_G = 0, opt_G_value = 0,
+	opt_g = 0,
+	opt_G = 0,
 	verbose = 0,				/* Verbose flag. */
 	debug = 0,
 	opt_version = 0,
@@ -130,9 +145,15 @@ unsigned int opt_a = 0,					/* Show-all-interfaces flag. */
 	opt_reset = 0,
 	opt_status = 0,
 	opt_watch = 0;
-static int nway_advertise = -1;
+static int msg_level = -1;
+static int set_BMCR = -1;
+static int nway_advertise = 0;
 static int fixed_speed = -1;
 static int override_phy = -1;
+char *opt_G_string = NULL;
+
+/* Internal values. */
+int new_ioctl_nums;
 int skfd = -1;					/* AF_INET socket for ioctl() calls.	*/
 struct ifreq ifr;
 
@@ -141,6 +162,7 @@ int show_basic_mii(long ioaddr, int phy_id);
 int mdio_read(int skfd, int phy_id, int location);
 void mdio_write(int skfd, int phy_id, int location, int value);
 static int parse_advertise(const char *capabilities);
+static void monitor_status(long ioaddr, int phy_id);
 
 
 int
@@ -148,15 +170,25 @@ main(int argc, char **argv)
 {
 	int c, errflag = 0;
 	char **spp, *ifname;
+    char *progname = rindex(argv[0], '/') ? rindex(argv[0], '/')+1 : argv[0];
 
-	while ((c = getopt_long(argc, argv, "aA:DfF:G:p:rRsvVw?", longopts, 0)) != EOF)
+	while ((c = getopt_long(argc, argv, shortopts, longopts, 0)) != EOF)
 		switch (c) {
 		case 'a': opt_a++; break;
-		case 'A': nway_advertise = parse_advertise(optarg); break;
+		case 'A': nway_advertise |= parse_advertise(optarg);
+			if (nway_advertise == -1) errflag++;
+			break;
+		case 'C': set_BMCR = strtoul(optarg, NULL, 16); break;
 		case 'D': debug++;			break;
 		case 'f': opt_f++; break;
-		case 'F': fixed_speed = parse_advertise(optarg); break;
-		case 'G': opt_G++; opt_G_value = atoi(optarg); break;
+		case 'F': fixed_speed = parse_advertise(optarg);
+			if (fixed_speed == -1) errflag++;
+			break;
+		case 'g': opt_g++; break;
+		case 'G': opt_G++; opt_G_string = strdup(optarg); break;
+		case 'm': opt_watch++; opt_status++; break;
+		case 'M': msg_level = strtoul(optarg, NULL, 0); break;
+		case 'h': fprintf(stderr, long_usage_msg, progname); return 0;
 		case 'p': override_phy = atoi(optarg); break;
 		case 'r': opt_restart++;	break;
 		case 'R': opt_reset++;		break;
@@ -164,10 +196,10 @@ main(int argc, char **argv)
 		case 'v': verbose++;		break;
 		case 'V': opt_version++;	break;
 		case 'w': opt_watch++;		break;
-		case '?': errflag++;
+		case '?': errflag++;		break;
 		}
 	if (errflag) {
-		fprintf(stderr, usage_msg);
+		fprintf(stderr, usage_msg, progname);
 		return 2;
 	}
 
@@ -177,7 +209,7 @@ main(int argc, char **argv)
 	/* Open a basic socket. */
 	if ((skfd = socket(AF_INET, SOCK_DGRAM,0)) < 0) {
 		perror("socket");
-		exit(-1);
+		return 1;
 	}
 
 	if (debug)
@@ -199,13 +231,29 @@ main(int argc, char **argv)
 		fprintf(stderr, "Using the default interface 'eth0'.\n");
 	}
 
-	/* Get the vitals from the interface. */
-	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
-	if (ioctl(skfd, SIOCGMIIPHY, &ifr) < 0) {
-		fprintf(stderr, "SIOCGMIIPHY on %s failed: %s\n", ifname,
-				strerror(errno));
-		(void) close(skfd);
-		return 1;
+	/* Verify that the interface supports the ioctl(), and if
+	   it is using the new or old SIOCGMIIPHY value (grrr...).
+	 */
+	{
+		u16 *data = (u16 *)(&ifr.ifr_data);
+
+		strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
+		data[0] = 0;
+
+		if (ioctl(skfd, 0x8947, &ifr) >= 0) {
+			new_ioctl_nums = 1;
+		} else if (ioctl(skfd, SIOCDEVPRIVATE, &ifr) >= 0) {
+			new_ioctl_nums = 0;
+		} else {
+			fprintf(stderr, "SIOCGMIIPHY on %s failed: %s\n", ifname,
+					strerror(errno));
+			(void) close(skfd);
+			return 1;
+		}
+		if (verbose)
+			printf("  Using the %s SIOCGMIIPHY value on PHY %d "
+				   "(BMCR 0x%4.4x).\n",
+				   new_ioctl_nums ? "new" : "old", data[0], data[3]);
 	}
 
 	do_one_xcvr(skfd);
@@ -217,6 +265,7 @@ main(int argc, char **argv)
 int do_one_xcvr(int skfd)
 {
 	u16 *data = (u16 *)(&ifr.ifr_data);
+	u32 *data32 = (u32 *)(&ifr.ifr_data);
 	unsigned phy_id = data[0];
 
 	if (override_phy >= 0) {
@@ -224,13 +273,62 @@ int do_one_xcvr(int skfd)
 		phy_id = override_phy;
 	}
 
+	if (opt_g || opt_G || msg_level >= 0) {
+		if (ioctl(skfd, SIOCGPARAMS, &ifr) < 0) {
+			fprintf(stderr, "SIOCGPARAMS on %s failed: %s\n", ifr.ifr_name,
+					strerror(errno));
+			return -1;
+		}
+	}
+	if (opt_g) {
+		int i;
+		printf("Driver general parameter settings:");
+		for (i = 0; i*sizeof(u32) < sizeof(ifr.ifr_ifru); i++) {
+			printf(" %d", data32[i]);
+		}
+		printf(".\n");
+	}
 	if (opt_G) {
-		/* Set an undefined general-purpose value.
-		   Modify this code section to do special hacks such as
-		   mdio_write(skfd, phy_id, 19, opt_G_value);
+		/* Set up to four arbitrary driver parameters from the -G parameter.
+		   The format is comma separated integers, with a missing element
+		   retaining the previous value.
 		*/
-		u32 *data = (u32 *)(&ifr.ifr_data);
-		data[0] = opt_G_value;
+		char *str = opt_G_string;
+		int i;
+		for (i = 0; str && i < 4; i++) {
+			char *endstr;
+			u32 newval = strtol(str, &endstr, 0);
+			if (debug)
+				printf(" parse string '%s'  value %d end '%s'.\n",
+					   str, newval, endstr);
+			if (str == endstr) {
+				if (endstr[0] == ',') /* No parameter */
+					str = endstr+1;
+				else {
+					fprintf(stderr, "Invalid driver parameter '%s'.\n", str);
+					str = index(str, ',');
+				}
+			} else if (endstr[0] == ',') {
+				data32[i] = newval;
+				str = endstr + 1;
+			} else if (endstr[0] == 0) {
+				data32[i] = newval;
+				break;
+			}
+		}
+		printf("Setting new driver general parameters:");
+		for (i = 0; i*sizeof(u32) < sizeof(ifr.ifr_ifru); i++) {
+			printf(" %d", data32[i]);
+		}
+		printf(".\n");
+		if (ioctl(skfd, SIOCSPARAMS, &ifr) < 0) {
+			fprintf(stderr, "SIOCSPARAMS on %s failed: %s\n", ifr.ifr_name,
+					strerror(errno));
+			return -1;
+		}
+	}
+	if (msg_level >= 0) {
+		data32[0] = msg_level;
 		if (ioctl(skfd, SIOCSPARAMS, &ifr) < 0) {
 			fprintf(stderr, "SIOCSPARAMS on %s failed: %s\n", ifr.ifr_name,
 					strerror(errno));
@@ -243,7 +341,7 @@ int do_one_xcvr(int skfd)
 		mdio_write(skfd, phy_id, 0, 0x8000);
 	}
 	/* Note: PHY addresses > 32 are pseudo-MII devices, usually built-in. */
-	if (phy_id < 64  &&  nway_advertise >= 0) {
+	if (phy_id < 64  &&  nway_advertise > 0) {
 		printf(" Setting the media capability advertisement register of "
 			   "PHY #%d to 0x%4.4x.\n", phy_id, nway_advertise | 1);
 		mdio_write(skfd, phy_id, 4, nway_advertise | 1);
@@ -260,12 +358,21 @@ int do_one_xcvr(int skfd)
 		int reg0_val = 0;
 		if (fixed_speed & 0x0180) 		/* 100mpbs */
 			reg0_val |=  0x2000;
-		if (fixed_speed & 0x0140)		/* Full duplex */
+		if ((fixed_speed & 0x0140) &&		/* A full duplex type and */
+			! (fixed_speed & 0x0820)) 		/* no half duplex types. */
 			reg0_val |= 0x0100;
 		printf("Setting the speed to \"fixed\", Control register %4.4x.\n",
 			   reg0_val);
 		mdio_write(skfd, phy_id, 0, reg0_val);
 	}
+	if (set_BMCR >= 0) {
+		printf("Setting the Basic Mode Control Register to 0x%4.4x.\n",
+			   set_BMCR);
+		mdio_write(skfd, phy_id, 0, set_BMCR);
+	}
+
+	if (opt_watch && opt_status)
+		monitor_status(skfd, phy_id);
 
 	show_basic_mii(skfd, phy_id);
 #ifdef LIBMII
@@ -299,7 +406,7 @@ int mdio_read(int skfd, int phy_id, int location)
 	data[0] = phy_id;
 	data[1] = location;
 
-	if (ioctl(skfd, SIOCGMIIREG, &ifr) < 0) {
+	if (ioctl(skfd, new_ioctl_nums ? 0x8948 : SIOCDEVPRIVATE+1, &ifr) < 0) {
 		fprintf(stderr, "SIOCGMIIREG on %s failed: %s\n", ifr.ifr_name,
 				strerror(errno));
 		return -1;
@@ -315,7 +422,7 @@ void mdio_write(int skfd, int phy_id, int location, int value)
 	data[1] = location;
 	data[2] = value;
 
-	if (ioctl(skfd, SIOCSMIIREG, &ifr) < 0) {
+	if (ioctl(skfd, new_ioctl_nums ? 0x8949 : SIOCDEVPRIVATE+2, &ifr) < 0) {
 		fprintf(stderr, "SIOCSMIIREG on %s failed: %s\n", ifr.ifr_name,
 				strerror(errno));
 	}
@@ -328,6 +435,7 @@ static int parse_advertise(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if ( ! capabilities) {
@@ -340,9 +448,13 @@ static int parse_advertise(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtol(capabilities, NULL, 16)) <= 0xffff)
+	if ((i = strtol(capabilities, &endptr, 16)) <= 0xffff  &&  endptr[0] == 0)
 		return i;
-	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
+	fprintf(stderr, "Invalid media advertisement value '%s'.\n"
+			"  Either pass a numeric value or one of the following names:\n",
+			capabilities);
+	for (i = 0; mtypes[i]; i++)
+		fprintf(stderr, "   %-14s %3.3x\n", mtypes[i], cap_map[i]);
 	return -1;
 }
 
@@ -376,9 +488,12 @@ int show_basic_mii(long ioaddr, int phy_id)
 		printf(".\n");
 	}
 
-	if (mii_val[0] == 0xffff) {
+	if (mii_val[0] == 0xffff  ||  mii_val[1] == 0x0000) {
 		printf("  No MII transceiver present!.\n");
-		return -1;
+		if (! opt_f) {
+			printf("  Use '--force' to view the information anyway.\n");
+			return -1;
+		}
 	}
 	/* Descriptive rename. */
 	bmcr = mii_val[0];
@@ -459,7 +574,7 @@ int show_basic_mii(long ioaddr, int phy_id)
 			   " autonegotiation).\n",
 			   lkpar & 0x0080 ? "100baseTx" : "10baseT");
 	else if ( ! (bmcr & 0x1000))
-		printf(" Link partner information information is not exchanged when in"
+		printf(" Link partner information is not exchanged when in"
 			   " fixed speed mode.\n");
 	else if ( ! (new_bmsr & 0x004))
 							;	/* If no partner, do not report status. */
@@ -470,17 +585,56 @@ int show_basic_mii(long ioaddr, int phy_id)
 	} else
 		printf(" Your link partner is strange, status %4.4x.\n", lkpar);
 
-	printf("   End of basic transceiver informaion.\n\n");
+	printf("   End of basic transceiver information.\n\n");
 	return 0;
 }
 
-#ifndef LIBMII
-extern void monitor_mii(long ioaddr, int phy_id)
+static void monitor_status(long ioaddr, int phy_id)
 {
-	fprintf(stderr, "\nThis version of 'mii-diag' has not been compiled with "
-			"the libmii.c library \n  required for the media monitor option.\n");
+	unsigned int baseline_1 = 0x55555555; 	/* Always show initial status. */
+
+	while (1) {
+		unsigned int new_1 = mdio_read(ioaddr, phy_id, 1);
+		if (new_1 != baseline_1) {
+			printf("%-12s 0x%4.4x 0x%4.4x\n",
+				   new_1 & 0x04 ? (new_1==0xffff ? "unknown" : "up") :
+				   new_1 & 0x20 ? "negotiating" : "down",
+				   new_1, mdio_read(ioaddr, phy_id, 5));
+			fflush(stdout);
+			baseline_1 = new_1;
+		}
+		sleep(1);
+	}
 }
-#endif /* not compiled for LIBMII */
+
+/* These functions are not used if we are linked with libmii. */
+
+int  show_mii_details(long ioaddr, int phy_id) __attribute__((weak));
+void monitor_mii(long ioaddr, int phy_id) __attribute__((weak));
+
+int  show_mii_details(long ioaddr, int phy_id)
+{
+	int mii_reg, mii_val;
+	printf(" MII PHY #%d transceiver registers:", phy_id);
+	for (mii_reg = 0; mii_reg < 32; mii_reg++) {
+		mii_val = mdio_read(skfd, phy_id, mii_reg);
+		printf("%s %4.4x", (mii_reg % 8) == 0 ? "\n  " : "",
+			   mii_val);
+	}
+	printf("\nThis version of 'mii-diag' has not been linked with "
+			"the libmii.c library.\n"
+			"  That library provides extended transceiver status reports.\n");
+	return 0;
+}
+
+void monitor_mii(long ioaddr, int phy_id)
+{
+	fprintf(stderr, "\nThis version of 'mii-diag' has not been linked with "
+			"the libmii.c library \n"
+			"  required for the media monitor option.\n");
+}
+
+
 
 /*
  * Local variables:
